@@ -26,6 +26,7 @@ __author__ = 'Sergey Konakov <skonakov@gmail.com>'
 
 import argparse
 import os
+import pkg_resources
 import re
 import sh
 import sys
@@ -63,31 +64,29 @@ config = AttrDict()
 
 
 def get_media_info(file):
-    result = AttrDict()
-    xmlIO  = StringIO()
+    xmlIO = StringIO()
     sh.mediainfo(
         '--Output=XML',
         '-f', file,
         _out=xmlIO
     )
-    mediainfo = MediaInfo(xmlIO.getvalue())
-    for track in mediainfo.tracks:
-        if track.track_type == 'Video':
-            result.video = track
-        elif track.track_type == 'Audio':
-            result.audio = track
-        elif track.track_type == 'Text':
-            result.text = track
-        elif track.track_type == 'General':
-            result.general = track
-        else:
-            pass
 
-    return result
+    info = MediaInfo(xmlIO.getvalue().encode('utf-8'))
+    tracks = []
+
+    for track in info.tracks:
+        type = track.track_type.lower()
+        if type == 'general':
+            general_info = track
+        else:
+            tracks.append(track)
+
+    return general_info, tracks
 
 
 class EncodingProgress:
     FRAMES_RE = re.compile(r'^frame=\s*(\d+) .*')
+
     def __init__(self, title, total_frames):
         widgets = [
             title,
@@ -104,7 +103,9 @@ class EncodingProgress:
         )
         self.pbar.start()
 
-    def process_ffmpeg_line(self, line):
+    def process_ffmpeg_line(self, line, stdin, process):
+        process._stderr.append(line)
+
         lines = line.splitlines()
         for l in lines:
             match = self.FRAMES_RE.match(l)
@@ -123,47 +124,90 @@ class EncodingProgress:
         self.pbar.finish()
 
 
-def convert(filename, args):
-    cache_file(filename)
-    info = get_media_info(filename)
-
-    if info.video.format == 'AVC':
+def get_video_opts(index, track):
+    if track.format == 'AVC':
         method = '1pass'
         video_opts = [
+            '-map', '0:%s' % index,
             '-codec:v', 'copy'
         ]
     else:
         method = '2pass'
         video_opts = [
-            '-b:v', str(info.video.bit_rate),
+            '-map', '0:%s' % index,
+            '-b:v', str(track.bit_rate),
             '-codec:v', 'libx264',
             '-profile:v', 'high',
             '-level', '4.1'
         ]
 
-    if info.audio.format == 'AAC':
+    return method, video_opts
+
+
+def get_audio_opts(index, track):
+    if track.format == 'AAC':
         audio_opts = [
-            '-codec:a', 'copy'
+            '-map', '0:%s' % index,
+            '-codec:a:%s' % index, 'copy'
         ]
     else:
-        audio_opts = config.audio_encoder_opts + [
-            '-b:a'
+        audio_opts = [
+            '-map', '0:%s' % index,
+            '-codec:a:%s' % index, 'libfaac',
+            '-b:a:%s' % index
         ]
-        if info.audio.channel_s >= 6:
+        if track.channel_s >= 6:
             audio_opts.append('320K')
         else:
             audio_opts.append('160K')
 
-    subtitle_opts = [
-        '-codec:s', 'copy'
+    return audio_opts
+
+
+def get_subtitle_opts(index, track):
+    return [
+        '-map', '0:%s' % index,
+        '-codec:s', 'mov_text'
     ]
 
+
+def convert(filename, args):
+    cache_file(filename)
+    general_info, tracks = get_media_info(filename)
+
+    method = None
+    video_opts = None
+    audio_opts = []
+    subtitle_opts = []
+    metadata_opts = [
+        '-map_metadata', '0'
+    ]
     input_ops = [
         '-i', '%s' % filename,
     ]
 
-    out_file_name = '%s.mp4' % info.general.file_name
-    out_path = os.path.join(info.general.folder_name, out_file_name)
+    for index, track in enumerate(tracks):
+        track_type = track.track_type.lower()
+        if track_type == 'video':
+            if method is not None:
+                raise Exception(
+                    "2mp4 currently doesn't support multiple video streams :("
+                )
+            method, video_opts = get_video_opts(index, track)
+            frame_count = track.frame_count
+            if frame_count is None:
+                frame_count = float(
+                    general_info.duration
+                ) / 1000 * float(
+                    track.original_frame_rate
+                )
+        elif track_type == 'audio':
+            audio_opts += get_audio_opts(index, track)
+        elif track_type == 'text':
+            subtitle_opts += get_subtitle_opts(index, track)
+
+    out_file_name = '%s.mp4' % general_info.file_name
+    out_path = os.path.join(general_info.folder_name, out_file_name)
     sys.stderr.write('Encoding %s -> %s\n' % (filename, out_file_name))
 
     if os.path.exists(out_path):
@@ -171,14 +215,15 @@ def convert(filename, args):
         return
 
     if method == '1pass':
-        opts = input_ops + video_opts + audio_opts + subtitle_opts + [
-            '-y',
-            out_path
-        ]
+        opts = input_ops + video_opts + audio_opts + \
+            subtitle_opts + metadata_opts + [
+                '-y',
+                out_path
+            ]
         if args.dry_run:
             print 'ffmpeg ' + ' '.join(opts)
         else:
-            progress = EncodingProgress('Pass 1 of 1:', info.video.frame_count)
+            progress = EncodingProgress('Pass 1 of 1:', frame_count)
             p = sh.ffmpeg(
                 *opts,
                 _err=progress.process_ffmpeg_line,
@@ -197,10 +242,7 @@ def convert(filename, args):
         if args.dry_run:
             print 'ffmpeg ' + ' '.join(opts)
         else:
-            pass1_progress = EncodingProgress(
-                'Pass 1 of 2: ',
-                info.video.frame_count
-            )
+            pass1_progress = EncodingProgress('Pass 1 of 2: ', frame_count)
             p = sh.ffmpeg(
                 *opts,
                 _err=pass1_progress.process_ffmpeg_line,
@@ -209,18 +251,16 @@ def convert(filename, args):
             p.wait()
             pass1_progress.finish()
 
-        opts = input_ops + video_opts + audio_opts + subtitle_opts + [
-            '-pass', '2',
-            '-y',
-            out_path
-        ]
+        opts = input_ops + video_opts + audio_opts + \
+            subtitle_opts + metadata_opts + [
+                '-pass', '2',
+                '-y',
+                out_path
+            ]
         if args.dry_run:
             print 'ffmpeg ' + ' '.join(opts)
         else:
-            pass2_progress = EncodingProgress(
-                'Pass 2 of 2: ',
-                info.video.frame_count
-            )
+            pass2_progress = EncodingProgress('Pass 2 of 2: ', frame_count)
             p = sh.ffmpeg(
                 *opts,
                 _err=pass2_progress.process_ffmpeg_line,
@@ -233,7 +273,9 @@ def convert(filename, args):
 def check_required_programs():
     # Check that mediainfo is installed
     if sh.which('mediainfo') is None:
-        print '%s: Cannot find mediainfo, please install before continuing.' % (
+        print (
+            '%s: Cannot find mediainfo, please install before continuing.'
+        ) % (
             PROG_NAME
         )
         exit(1)
@@ -310,8 +352,14 @@ def main():
             "just show the command(s) that would be executed"
         )
     )
+    parser.add_argument(
+        '-v', '--version',
+        action='version',
+        version='2mp4 ' + pkg_resources.get_distribution('2mp4').version
+    )
 
     args = parser.parse_args()
+
     input_file = os.path.abspath(args.input_file.strip())
     if not os.path.exists(input_file):
         print '%s: %s: No such file or directory' % (PROG_NAME, input_file)
@@ -321,11 +369,21 @@ def main():
 
     os.chdir(tempfile.gettempdir())
 
-    if os.path.isfile(input_file):
-        convert(input_file, args)
-    else:
-        for file in os.listdir(input_file):
-            name, ext = os.path.splitext(file)
-            file = os.path.join(input_file, file)
-            if os.path.isfile(file) and ext.lower() in VIDEO_EXTENSIONS:
-                convert(file, args)
+    try:
+        if os.path.isfile(input_file):
+            convert(input_file, args)
+        else:
+            for file in os.listdir(input_file):
+                name, ext = os.path.splitext(file)
+                file = os.path.join(input_file, file)
+                if os.path.isfile(file) and ext.lower() in VIDEO_EXTENSIONS:
+                    convert(file, args)
+    except sh.ErrorReturnCode as e:
+        print
+        print
+        print "OUPS that wasn't supposed to happen!"
+        print
+        print 'STDERR:'
+        for line in StringIO(e.stderr):
+            print '\t', line,
+        exit(1)
